package ru.dbotthepony.mc.otm.capability.matter

import com.mojang.datafixers.Products
import com.mojang.serialization.Codec
import com.mojang.serialization.codecs.RecordCodecBuilder
import net.minecraft.nbt.CompoundTag
import net.minecraft.nbt.NbtOps
import net.minecraft.nbt.Tag
import net.minecraft.network.FriendlyByteBuf
import net.minecraft.resources.ResourceLocation
import net.minecraft.world.item.Item
import net.minecraft.world.item.ItemStack
import net.minecraft.world.item.Items
import net.minecraftforge.registries.ForgeRegistries
import ru.dbotthepony.mc.otm.core.*
import ru.dbotthepony.mc.otm.core.nbt.contains
import ru.dbotthepony.mc.otm.core.nbt.set
import ru.dbotthepony.mc.otm.data.UUIDCodec
import java.util.*

sealed interface IPatternState {
	val id: UUID
	val item: Item
	val researchPercent: Double

	fun asMutable(): MutablePatternState
	fun asImmutable() : PatternState

	fun matchId(other: IPatternState): Boolean {
		return other.id == id
	}

	fun copyAsMutable(
		id: UUID = this.id,
		item: Item = this.item,
		researchPercent: Double = this.researchPercent,
	): MutablePatternState {
		return MutablePatternState(id, item, researchPercent)
	}

	fun copyAsImmutable(
		id: UUID = this.id,
		item: Item = this.item,
		researchPercent: Double = this.researchPercent,
	) : PatternState {
		return PatternState(id, item, researchPercent)
	}

	fun stack(count: Int = 1): ItemStack {
		return ItemStack(item, count)
	}

	fun serializeNBT(): CompoundTag

	fun write(buff: FriendlyByteBuf) {
		buff.writeUUID(id)
		buff.writeItemType(item)
		buff.writeDouble(researchPercent)
	}
}

private fun <T : IPatternState> codec(builder: RecordCodecBuilder.Instance<T>): Products.P3<RecordCodecBuilder.Mu<T>, UUID, Item, Double> {
	return builder.group(
		UUIDCodec.fieldOf("id").forGetter(IPatternState::id),
		ForgeRegistries.ITEMS.codec.fieldOf("item").forGetter(IPatternState::item),
		Codec.doubleRange(0.0, 1.0).fieldOf("researchPercent").forGetter(IPatternState::researchPercent)
	)
}

private fun read(buff: FriendlyByteBuf, mutable: Boolean): IPatternState? {
	val id = buff.readUUID()
	val item = buff.readItemType()
	val researchPercent = buff.readDouble()

	item ?: return null

	if (mutable) {
		return MutablePatternState(id, item, researchPercent)
	} else {
		return PatternState(id, item, researchPercent)
	}
}

data class PatternState(
	override val id: UUID,
	override val item: Item,
	override val researchPercent: Double,
) : IPatternState {
	override fun asMutable(): MutablePatternState {
		return MutablePatternState(id, item, researchPercent)
	}

	override fun asImmutable(): PatternState {
		return this
	}

	override fun serializeNBT(): CompoundTag {
		return CODEC.encode(this, NbtOps.INSTANCE, NbtOps.INSTANCE.empty()).get().map({ it as CompoundTag }, { throw RuntimeException("Failed to serialize PatternState: ${it.message()}") })
	}

	companion object {
		fun deserializeNBT(tag: Tag?): PatternState? {
			tag ?: return null
			return CODEC.decode(NbtOps.INSTANCE, tag).result().map { it.first }.orElse(null)
		}

		fun read(buff: FriendlyByteBuf): PatternState? {
			return read(buff, false) as PatternState?
		}

		val CODEC: Codec<PatternState> by lazy {
			RecordCodecBuilder.create {
				codec(it).apply(it, ::PatternState)
			}
		}
	}
}

data class MutablePatternState(
	override val id: UUID,
	override val item: Item,
	override var researchPercent: Double,
) : IPatternState {
	override fun asMutable(): MutablePatternState {
		return this
	}

	override fun asImmutable(): PatternState {
		return PatternState(id, item, researchPercent)
	}

	override fun serializeNBT(): CompoundTag {
		return CODEC.encode(this, NbtOps.INSTANCE, NbtOps.INSTANCE.empty()).get().map({ it as CompoundTag }, { throw RuntimeException("Failed to serialize MutablePatternState: ${it.message()}") })
	}

	companion object {
		fun deserializeNBT(tag: Tag?): MutablePatternState? {
			tag ?: return null
			return CODEC.decode(NbtOps.INSTANCE, tag).result().map { it.first }.orElse(null)
		}

		fun read(buff: FriendlyByteBuf): MutablePatternState? {
			return read(buff, true) as MutablePatternState?
		}

		val CODEC: Codec<MutablePatternState> by lazy {
			RecordCodecBuilder.create {
				codec(it).apply(it, ::MutablePatternState)
			}
		}
	}
}
