package ru.dbotthepony.mc.otm.capability.matter

import com.mojang.datafixers.Products
import com.mojang.serialization.Codec
import com.mojang.serialization.codecs.RecordCodecBuilder
import net.minecraft.nbt.CompoundTag
import net.minecraft.nbt.NbtOps
import net.minecraft.nbt.Tag
import net.minecraft.network.FriendlyByteBuf
import net.minecraft.resources.ResourceLocation
import net.minecraft.world.item.Item
import net.minecraft.world.item.ItemStack
import net.minecraft.world.item.Items
import net.minecraftforge.registries.ForgeRegistries
import ru.dbotthepony.mc.otm.core.readItemType
import ru.dbotthepony.mc.otm.core.writeItemType
import ru.dbotthepony.mc.otm.data.UUIDCodec
import java.util.Optional
import java.util.UUID

sealed interface IReplicationTask<S : IReplicationTask<S>> {
	val id: UUID
	val patternId: Optional<UUID>
	val item: Item
	val inProgress: Int
	val finished: Int
	val required: Int

	fun asMutable(): MutableReplicationTask
	fun asImmutable(): ReplicationTask

	val total get() = inProgress + finished + required

	fun copyAsMutable(
		id: UUID = this.id,
		patternId: Optional<UUID> = this.patternId,
		item: Item = this.item,
		inProgress: Int = this.inProgress,
		finished: Int = this.finished,
		required: Int = this.required,
	): MutableReplicationTask {
		return MutableReplicationTask(id, patternId, item, inProgress, finished, required)
	}

	fun copyAsImmutable(
		id: UUID = this.id,
		patternId: Optional<UUID> = this.patternId,
		item: Item = this.item,
		inProgress: Int = this.inProgress,
		finished: Int = this.finished,
		required: Int = this.required,
	): ReplicationTask {
		return ReplicationTask(id, patternId, item, inProgress, finished, required)
	}

	fun matchId(other: IReplicationTask<*>): Boolean {
		return other.id == id
	}

	fun stack(count: Int = 1): ItemStack {
		return ItemStack(item, count)
	}

	fun allocate(amount: Int = 1): S
	fun finish(amount: Int = 1): S

	fun serializeNBT(): CompoundTag

	fun write(buff: FriendlyByteBuf) {
		buff.writeUUID(id)

		buff.writeBoolean(patternId.isPresent)
		patternId.ifPresent(buff::writeUUID)

		buff.writeItemType(item)

		buff.writeInt(inProgress)
		buff.writeInt(finished)
		buff.writeInt(required)
	}
}

private fun <T : IReplicationTask<T>> codec(builder: RecordCodecBuilder.Instance<T>): Products.P6<RecordCodecBuilder.Mu<T>, UUID, Optional<UUID>, Item, Int, Int, Int> {
	return builder.group(
		UUIDCodec.fieldOf("id").forGetter(IReplicationTask<*>::id),
		UUIDCodec.optionalFieldOf("patternId").forGetter(IReplicationTask<*>::patternId),
		ForgeRegistries.ITEMS.codec.fieldOf("item").forGetter(IReplicationTask<*>::item),
		Codec.intRange(0, Int.MAX_VALUE).fieldOf("inProgress").forGetter(IReplicationTask<*>::inProgress),
		Codec.intRange(0, Int.MAX_VALUE).fieldOf("finished").forGetter(IReplicationTask<*>::finished),
		Codec.intRange(0, Int.MAX_VALUE).fieldOf("required").forGetter(IReplicationTask<*>::required),
	)
}

private fun read(buff: FriendlyByteBuf, mutable: Boolean): IReplicationTask<*>? {
	val id = buff.readUUID()
	val patternId: UUID? = if (buff.readBoolean()) buff.readUUID() else null
	val item = buff.readItemType()
	val inProgress = buff.readInt()
	val finished = buff.readInt()
	val required = buff.readInt()

	item ?: return null

	if (mutable) {
		return MutableReplicationTask(id, Optional.ofNullable(patternId), item, inProgress, finished, required)
	} else {
		return ReplicationTask(id, Optional.ofNullable(patternId), item, inProgress, finished, required)
	}
}

data class ReplicationTask(
	override val id: UUID,
	override val patternId: Optional<UUID>,
	override val item: Item,
	override val inProgress: Int,
	override val finished: Int,
	override val required: Int
) : IReplicationTask<ReplicationTask> {
	override fun asMutable(): MutableReplicationTask {
		return MutableReplicationTask(id, patternId, item, inProgress, finished, required)
	}

	override fun asImmutable(): ReplicationTask {
		return this
	}

	override fun allocate(amount: Int): ReplicationTask {
		return ReplicationTask(id, patternId, item, inProgress + amount, finished, required - amount)
	}

	override fun finish(amount: Int): ReplicationTask {
		return ReplicationTask(id, patternId, item, inProgress - amount, finished + amount, required)
	}

	override fun serializeNBT(): CompoundTag {
		return CODEC.encode(this, NbtOps.INSTANCE, NbtOps.INSTANCE.empty()).get().map({ it as CompoundTag }, { throw RuntimeException("Failed to serialize ReplicationTask: ${it.message()}") })
	}

	companion object {
		fun deserializeNBT(nbt: Tag?): ReplicationTask? {
			nbt ?: return null
			return CODEC.decode(NbtOps.INSTANCE, nbt).result().map { it.first }.orElse(null)
		}

		fun read(buff: FriendlyByteBuf): ReplicationTask? {
			return read(buff, false) as ReplicationTask?
		}

		val CODEC: Codec<ReplicationTask> by lazy {
			RecordCodecBuilder.create {
				codec(it).apply(it, ::ReplicationTask)
			}
		}
	}
}

data class MutableReplicationTask(
	override val id: UUID,
	override val patternId: Optional<UUID>,
	override val item: Item,
	override var inProgress: Int,
	override var finished: Int,
	override var required: Int
) : IReplicationTask<MutableReplicationTask> {
	override fun asMutable(): MutableReplicationTask {
		return this
	}

	override fun asImmutable(): ReplicationTask {
		return ReplicationTask(id, patternId, item, inProgress, finished, required)
	}

	override fun allocate(amount: Int): MutableReplicationTask {
		inProgress += amount
		finished -= amount
		return this
	}

	override fun finish(amount: Int): MutableReplicationTask {
		finished += amount
		inProgress -= amount
		return this
	}

	override fun serializeNBT(): CompoundTag {
		return CODEC.encode(this, NbtOps.INSTANCE, NbtOps.INSTANCE.empty()).get().map({ it as CompoundTag }, { throw RuntimeException("Failed to serialize MutableReplicationTask: ${it.message()}") })
	}

	companion object {
		fun deserializeNBT(nbt: Tag?): MutableReplicationTask? {
			nbt ?: return null
			return CODEC.decode(NbtOps.INSTANCE, nbt).result().map { it.first }.orElse(null)
		}

		fun read(buff: FriendlyByteBuf): MutableReplicationTask? {
			return read(buff, true) as MutableReplicationTask?
		}

		val CODEC: Codec<MutableReplicationTask> by lazy {
			RecordCodecBuilder.create {
				codec(it).apply(it, ::MutableReplicationTask)
			}
		}
	}
}
